/**
 * @fileoverview To show error when followed bad practices inside mapStateToProps of react redux applications
 * @author Sivanesh
 */
'use strict';

//------------------------------------------------------------------------------
// Rule Definition
//------------------------------------------------------------------------------

function isInReturn(node) {
  let n = node;

  const returnBodyCondn = n =>
    n.type === 'ArrowFunctionExpression' &&
    n.body &&
    n.body.type === 'ObjectExpression';

  while (n && !(n.type === 'ReturnStatement' || returnBodyCondn(n))) {
    n = n.parent;
  }

  if (n && n.type === 'ReturnStatement') {
    return true;
  } else if (n && returnBodyCondn(n)) {
    // const mapStateToProps = (state) => ({})
    if (
      n.parent.type === 'VariableDeclarator' &&
      n.parent.id.name === 'mapStateToProps'
    ) {
      return true;
    } else if (
      n.parent.type === 'CallExpression' &&
      n.parent.callee.name === 'connect'
    ) {
      // connect((state) => ({})
      return true;
    }
  } else {
    return false;
  }
}

module.exports = {
  meta: {
    docs: {
      description:
        'To show error when followed bad practices inside mapStateToProps of react redux applications',
      category: 'Performance',
      recommended: true
    },
    fixable: null, // or "code" or "whitespace"
    schema: [
      // fill in your schema
    ]
  },

  create: function(context) {
    const vulnerableQuery = ':matches(ArrayExpression, ObjectExpression)';
    const idNameQuery = '[id.name="mapStateToProps"]';
    const connectCallQuery = 'CallExpression[callee.name="connect"]';

    //----------------------------------------------------------------------
    // Helpers
    //----------------------------------------------------------------------

    function foundInitialization(node) {
      // Switching the process as inside return and outside return.
      if (!isInReturn(node)) {
        return;
      }

      // For All return statement's Objects.
      if (node.parent.type === 'ReturnStatement') return;

      // connect(state => ({}))
      if (
        node.parent.type === 'ArrowFunctionExpression' &&
        node.parent.parent &&
        node.parent.parent.callee &&
        node.parent.parent.callee.name === 'connect'
      ) {
        return;
      }

      // const mapStateToProps = state =>  ({name: 1})
      if (
        node.type === 'ObjectExpression' &&
        node.parent.type === 'ArrowFunctionExpression' &&
        node.parent &&
        node.parent.parent &&
        node.parent.parent.id &&
        node.parent.parent.id.name === 'mapStateToProps'
      ) {
        return;
      }

      // If return objects function call has any array or object
      if (node.parent.type === 'CallExpression') {
        return;
      }

      // const { name = {} } = state || {}
      if (
        node.parent.type === 'LogicalExpression' &&
        node.parent.parent &&
        node.parent.parent.id &&
        node.parent.parent.id.type === 'ObjectPattern'
      ) {
        return;
      }

      // const { name = {} } = state ? state:  {}
      if (
        node.parent.type === 'ConditionalExpression' &&
        node.parent.parent &&
        node.parent.parent.id &&
        node.parent.parent.id.type === 'ObjectPattern'
      ) {
        return;
      }

      context.report({
        node,
        message:
          'Causes performance issue as it creates a new reference each time when redux calls this connect.'
      });
    }

    const arrRefs = [
      'concat',
      'from',
      'of',
      'filter',
      'flat',
      'flatMap',
      'keys',
      'map',
      'slice',
      'values'
    ];

    const objRefs = ['assign', 'create'];

    function mstpFunction(node) {
      if (!isInReturn(node)) {
        return;
      }

      if (node.callee.property && arrRefs.includes(node.callee.property.name)) {
        context.report({
          node,
          message:
            'Creates new Array reference. Hence causes performance issues.'
        });
      }

      if (node.callee.object && node.callee.object.name === 'Object') {
        if (
          node.callee.property &&
          objRefs.includes(node.callee.property.name)
        ) {
          context.report({
            node,
            message:
              'Creates new Object reference. Hence causes performance issues.'
          });
        }
      }
    }

    // add a function that takes a list of files it obtains mapStateToProps and count.

    //----------------------------------------------------------------------
    // Public
    //----------------------------------------------------------------------

    return {
      [`FunctionDeclaration${idNameQuery} ${vulnerableQuery}`]: foundInitialization,
      [`VariableDeclarator${idNameQuery} ArrowFunctionExpression ${vulnerableQuery}`]: foundInitialization,
      [`${connectCallQuery} ArrowFunctionExpression ${vulnerableQuery}`]: foundInitialization,
      [`${connectCallQuery} FunctionExpression ${vulnerableQuery}`]: foundInitialization,

      // functions check
      [`FunctionDeclaration${idNameQuery} CallExpression`]: mstpFunction,
      [`VariableDeclarator${idNameQuery} ArrowFunctionExpression CallExpression`]: mstpFunction,
      [`${connectCallQuery} ArrowFunctionExpression CallExpression`]: mstpFunction,
      [`${connectCallQuery} FunctionExpression CallExpression`]: mstpFunction
    };
  }
};
